package com.psmreborn.nopsmdrm;

import static com.psmreborn.nopsmdrm.Helper.*;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.ProgressDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.pm.PackageManager;
import android.os.AsyncTask;
import android.os.Build;
import android.os.Environment;
import android.telephony.TelephonyManager;
import android.util.Log;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;

import eu.chainfire.libsuperuser.Shell;

public class Installer extends AsyncTask<Void, Void, Void> {
    private boolean wasError = false;
    private String errorMsg = "";

    private ProgressDialog dialog = null;
    private Context ctx = null;
    private StringEncryptor stringEncryptor = null;
    private String busyboxBinary = "";
    public Installer(Context context){
        this.ctx = context;
        this.stringEncryptor = new StringEncryptor(this.ctx);
    }

    private void setError(String msg) throws Exception {
        this.wasError = true;
        this.errorMsg = msg;
        Log.e("INSTALLER",errorMsg);
        throw new Exception(msg);
    }
    private static void writeTxtFile(String txtFile, String txt) throws IOException {
        Log.i("INSTALLER", "Writing: "+txtFile);
        FileWriter txtStream = new FileWriter(txtFile);
        txtStream.write(txt);
        txtStream.close();
    }
    private void copyTo(InputStream inpStream, OutputStream outStream) throws IOException {
        int totalRead = 0;
        byte[] buffer = new byte[0x1000];

        do {
            totalRead = inpStream.read(buffer, 0, buffer.length);
            outStream.write(buffer, 0, totalRead);
        }
        while(totalRead >= buffer.length);

        outStream.flush();
    }

    private void unpackResource(int resourceId, File outputFile) throws IOException {
        InputStream resourceStream = ctx.getResources().openRawResource(resourceId);

        FileOutputStream fs = new FileOutputStream(outputFile);
        copyTo(resourceStream, fs);
        fs.close();

        resourceStream.close();
    }

    private boolean fileExistRoot(String filename) throws Shell.ShellDiedException {
        int res = Shell.Pool.SU.run(new String[] { this.busyboxBinary + " stat '" + filename +"'" });
        return res == 0;
    }

    private void moveFileRoot(String filename, String destFilename) throws Exception {
        int res = Shell.Pool.SU.run(new String[] { this.busyboxBinary +" mv '"+ filename + "' '" +destFilename +"'"});
        if(res != 0){
            this.setError("Failed to rename "+filename+" to "+ destFilename);
        }
    }
    private void mkdirAndChmodChown(String directory, int chmod, String chown) throws Shell.ShellDiedException {
        Shell.Pool.SU.run(new String[]{
                this.busyboxBinary + " mkdir '" + directory +"'",
                this.busyboxBinary + " chmod " + String.valueOf(chmod) +" '"+directory+"'",
                this.busyboxBinary + " chown " + chown +":"+chown+" '"+directory+"'"
        });
    }
    private void copyChmodAndChown(String srcFile, String dstFile, int chmod, String chown) throws Exception {
        int res = Shell.Pool.SU.run(new String[]{
                this.busyboxBinary + " cp '" + srcFile +"' '"+dstFile+"'",
                this.busyboxBinary + " chmod " + String.valueOf(chmod) +" '"+dstFile+"'",
                this.busyboxBinary + " chown " + chown +":"+chown+" '"+dstFile+"'"
        });

        if(res != 0){
            this.setError("Failed to copy & change mode.");
            return;
        }
    }


    private void setupBusyBox() throws Exception {
        Log.i("INSTALLER","Creating busybox binary");
        File tmpFile = new File(ctx.getCacheDir(), "busybox");
        tmpFile.createNewFile();

        if(tmpFile.setExecutable(true,false)) {
            unpackResource(R.raw.busybox, tmpFile);
            this.busyboxBinary = tmpFile.getAbsolutePath();
        }
        else {
            this.setError("failed to extract busybox binary.");
            return;
        }

    }


    private void generateDeviceFingerprint(String outputFilename) throws FileNotFoundException, UnsupportedEncodingException {


        TelephonyManager tm = ((TelephonyManager) ctx.getSystemService( Context.TELEPHONY_SERVICE));

        String deviceId = "(blank)";
        if(tm != null)
            deviceId = tm.getDeviceId();

        if(deviceId == null)
            deviceId = "(blank)";

        String serial = Build.SERIAL;
        if(serial == null)
            serial = "(blank)";

        String brand = Build.BRAND;
        if(brand == null)
            brand = "(blank)";

        String manu = Build.MANUFACTURER;
        if(manu == null)
            manu = "(blank)";

        String model = Build.MODEL;
        if(model == null)
            model = "(blank)";

        String product = Build.PRODUCT;
        if(product == null)
            product = "(blank)";

        String device = Build.DEVICE;
        if(device == null)
            device = "(blank)";

        String type = Build.TYPE;
        if(type == null)
            type = "(blank)";

        PrintWriter writer = new PrintWriter(outputFilename, "UTF-8");
        writer.println(String.valueOf(stringEncryptor.getPsmUid()));
        writer.println(stringEncryptor.getAndroidId());
        writer.println(deviceId);
        writer.println(serial);
        writer.println(brand);
        writer.println(manu);
        writer.println(model);
        writer.println(product);
        writer.println(device);
        writer.println(type);
        writer.close();
    }

    private void backupPsm() throws Exception {
        String psmSdcardFolder = new File(Environment.getExternalStorageDirectory(), "psm").getAbsolutePath();
        new File(psmSdcardFolder).mkdirs();
        String filename = "psm_"+getDateTime();

        int res = Shell.Pool.SU.run(new String[]{this.busyboxBinary + " tar c '" + getPsmApp().dataDir + "' -f '" + new File(psmSdcardFolder, filename+".tar").getAbsolutePath()+"'"});
        if (res != 0) {
            this.setError("Failed to backup existing PSM data. (exit code: " + res + ")");
            return;
        }
        generateDeviceFingerprint(new File(psmSdcardFolder, filename + "_DEV.txt").getAbsolutePath());
    }

    private void makeDirs() throws PackageManager.NameNotFoundException, Shell.ShellDiedException {
        mkdirAndChmodChown(new File(getPsmApp().dataDir, "shared_prefs").getAbsolutePath(), 771, String.valueOf(stringEncryptor.getPsmUid()));
        mkdirAndChmodChown(new File(getPsmApp().dataDir, "files").getAbsolutePath(), 771, String.valueOf(stringEncryptor.getPsmUid()));
        mkdirAndChmodChown(new File(new File(getPsmApp().dataDir, "files"), "kdc").getAbsolutePath(), 771, String.valueOf(stringEncryptor.getPsmUid()));
        mkdirAndChmodChown(new File(getPsmApp().dataDir, "databases").getAbsolutePath(), 771, String.valueOf(stringEncryptor.getPsmUid()));
    }

    private void patchSharedPrefs() throws Exception {
        // get the path to the shared_prefs folder
        String sharedPrefsPath = new File(getPsmApp().dataDir, "shared_prefs").getAbsolutePath();
        // check if signininfo.xml file exists or not ...
        boolean signinInfoExist = fileExistRoot(new File(sharedPrefsPath, "SigninInfo.xml").getAbsolutePath());

        if (!signinInfoExist) { // if file not found ...
                          // then generate our own shared_prefs

            // encrypt the actual strings, username, password, etc
            String emailAddress = stringEncryptor.encryptString("nopsmdrm@transrights.lgbt");
            String password = stringEncryptor.encryptString("password");
            String accountId = stringEncryptor.encryptString(String.valueOf(0x123456789ABCDEFL));

            // get the cache folder for our 'shared_prefs'
            String tmpPrefsFolder = ctx.getCacheDir().getAbsolutePath();

            // work out paths to each file ...
            String csigninInfo = new File(tmpPrefsFolder, "SigninInfo.xml").getAbsolutePath();
            String cpsstorePrefs = new File(tmpPrefsFolder, "com.playstation.psstore_preferences.xml").getAbsolutePath();
            String crunningContentInfo = new File(tmpPrefsFolder, "RunningContentInfo.xml").getAbsolutePath();
            String clocalLibrary = new File(tmpPrefsFolder, "LocalLibrary.xml").getAbsolutePath();

            String rsigninInfo = new File(sharedPrefsPath, "SigninInfo.xml").getAbsolutePath();
            String rpsstorePrefs = new File(sharedPrefsPath, "com.playstation.psstore_preferences.xml").getAbsolutePath();
            String rrunningContentInfo = new File(sharedPrefsPath, "RunningContentInfo.xml").getAbsolutePath();
            String rlocalLibrary = new File(sharedPrefsPath, "LocalLibrary.xml").getAbsolutePath();

            // generate shared_prefs
            writeTxtFile(csigninInfo, "<?xml version='1.0' encoding='utf-8' standalone='yes' ?>\n<map>\n<string name=\"SignedInUsername\">"+emailAddress+"</string>\n<boolean name=\"PassSave\" value=\"true\" />\n<string name=\"Password\">"+password+"</string>\n<boolean name=\"AutoSignIn\" value=\"true\" />\n</map>\n");
            writeTxtFile(cpsstorePrefs, "<?xml version='1.0' encoding='utf-8' standalone='yes' ?>\n<map>\n<boolean name=\"key_upgradeDownloadTableForNeedWifi\" value=\"true\" />\n<string name=\"last_signin_account_id\">"+accountId+"</string>\n<long name=\"last_signin_account_region\" value=\"2\" />\n<int name=\"key_psstore\" value=\"1\" />\n<int name=\"key_downloader\" value=\"1\" />\n<int name=\"psm_license_agree_version_code\" value=\"1170\" />\n<boolean name=\"key_notDisplayAgainEndOfSupportPreNavi\" value=\"true\" />\n<int name=\"key_xmlcache\" value=\"1\" />\n<string name=\"last_signin_account_country\">US</string>\n<boolean name=\"key_notDisplayAgainContentStartNavi\" value=\"true\" />\n<int name=\"key_startcontent\" value=\"1\" />\n<int name=\"key_nsxevent\" value=\"1\" />\n<boolean name=\"key_upgradeLibraryTableForLocationUseConfirmationDate\" value=\"true\" />\n<int name=\"key_install\" value=\"1\" />\n<string name=\"update_md5\">387ce7e424258aef426aaa5be8a1638a</string>\n<boolean name=\"psm_license_agree\" value=\"true\" />\n<int name=\"key_guestinfo\" value=\"1\" />\n<string name=\"last_signin_account_language\">en</string>\n<int name=\"key_cache\" value=\"2\" />\n<int name=\"key_signinfo\" value=\"2\" />\n</map>\n");
            writeTxtFile(crunningContentInfo, "<?xml version='1.0' encoding='utf-8' standalone='yes' ?>\n<map>\n<null name=\"title_id\" />\n<null name=\"next_title_id\" />\n</map>\n");
            writeTxtFile(clocalLibrary, "<?xml version='1.0' encoding='utf-8' standalone='yes' ?>\n<map>\n<boolean name=\"notDisplayAgain\" value=\"true\" />\n<int name=\"sortType\" value=\"0\" />\n<boolean name=\"isList\" value=\"false\" />\n</map>\n");

            // copy to the correct place and set permissions properly.
            copyChmodAndChown(csigninInfo, rsigninInfo, 660, String.valueOf(stringEncryptor.getPsmUid()));
            copyChmodAndChown(cpsstorePrefs, rpsstorePrefs, 660, String.valueOf(stringEncryptor.getPsmUid()));
            copyChmodAndChown(crunningContentInfo, rrunningContentInfo, 660, String.valueOf(stringEncryptor.getPsmUid()));
            copyChmodAndChown(clocalLibrary, rlocalLibrary, 660, String.valueOf(stringEncryptor.getPsmUid()));

        }

    }

    private void installNoPsmDrmModules() throws Exception {
        String nativeLibsFolder = getPsmApp().nativeLibraryDir;

        String rlibPsmKdcFile = new File(nativeLibsFolder, "libpsmkdc_jni.so").getAbsolutePath();
        String rlibDefaultFile = new File(nativeLibsFolder, "libdefault.so").getAbsolutePath();
        String realLibDefaultFile = new File(nativeLibsFolder, "libdefault_real.so").getAbsolutePath();
        boolean realLibDefaultExist = fileExistRoot(realLibDefaultFile);

        if(!realLibDefaultExist) {
            // if libdefault_real.so not found, then rename libdefault.so to libdefault_real.so ...
            moveFileRoot(rlibDefaultFile, realLibDefaultFile);
        }

        String clibDefault = new File(ctx.getCacheDir(), "libdefault.so").getAbsolutePath();
        String clibPsmKdc = new File(ctx.getCacheDir(), "libpsmkdc_jni.so").getAbsolutePath();

        // unpack the library files ...
        unpackResource(R.raw.libdefault, new File(clibDefault));
        unpackResource(R.raw.libpsmkdc_jni, new File(clibPsmKdc));

        copyChmodAndChown(clibDefault, rlibDefaultFile, 755, "system");
        copyChmodAndChown(clibPsmKdc, rlibPsmKdcFile, 755, "system");

    }
    private void installDatabase() throws Exception {
        String libraryDbFile = new File(ctx.getCacheDir(), "library.db").getAbsolutePath();

        String databasesFolder = new File(getPsmApp().dataDir, "databases").getAbsolutePath();
        String rlibraryDbFile = new File(databasesFolder, "library.db").getAbsolutePath();

        unpackResource(R.raw.library, new File(libraryDbFile));
        copyChmodAndChown(libraryDbFile, rlibraryDbFile, 660, String.valueOf(stringEncryptor.getPsmUid()));
    }

    @Override
    protected void onPreExecute() {
        Shell.setRedirectDeprecated(false);

        dialog = new ProgressDialog(ctx);
        dialog.setTitle("Installing NoPsmDrm ...");
        dialog.setMessage("Please Wait ...");
        dialog.setIndeterminate(true);
        dialog.setCancelable(false);
        dialog.show();
    }
    @Override
    protected Void doInBackground(Void... voids) {

        try {
            setupBusyBox();

            if (isPsmInstalled()) {
                backupPsm();
                makeDirs();
                patchSharedPrefs();
                installNoPsmDrmModules();
                installDatabase();
            }
        }
        catch(Exception e){
            this.wasError = true;
            this.errorMsg = e.getMessage();
            return null;
        }

        return null;
    }

    @Override
    protected void onPostExecute(Void result) {
        dialog.dismiss();

        if(wasError) {
            new AlertDialog.Builder((Activity)ctx)
                    .setTitle("Error Occurred.")
                    .setMessage(this.errorMsg)
                    .setCancelable(false)
                    .setPositiveButton("Ok", new DialogInterface.OnClickListener() {
                        @Override
                        public void onClick(DialogInterface dialog, int which) {

                        }
                    }).show();
        }
        else{
            new AlertDialog.Builder((Activity)ctx)
                    .setTitle("Installed!")
                    .setMessage("Your PSM Application was patched successfully!\nNote: WI-FI has to be turned off for games to work.")
                    .setCancelable(false)
                    .setPositiveButton("Ok", new DialogInterface.OnClickListener() {
                        @Override
                        public void onClick(DialogInterface dialog, int which) {

                        }
                    }).show();
        }
    }
}
